mirror of https://lore.kernel.org/lkml/
 help / color / mirror / Atom feed
* [PATCH] sched: do not call __put_task_struct() on rt if pi_blocked_on is set
@ 2025-04-09  0:27 Luis Claudio R. Goncalves
  2025-04-09 15:43 ` Steven Rostedt
  0 siblings, 1 reply; 5+ messages in thread
From: Luis Claudio R. Goncalves @ 2025-04-09  0:27 UTC (permalink / raw)
  To: Sebastian Andrzej Siewior, Clark Williams, Steven Rostedt,
	Tejun Heo, David Vernet, Barret Rhoden, Josh Don, Crystal Wood,
	Luis Claudio R. Goncalves, linux-kernel, linux-rt-devel,
	Juri Lelli, lclaudio00

With PREEMPT_RT enabled, some of the calls to put_task_struct() coming
from rt_mutex_adjust_prio_chain() could happen in preemptible context and
with a mutex enqueued. That could lead to this sequence:

	rt_mutex_adjust_prio_chain()
	  put_task_struct()
	    __put_task_struct()
	      sched_ext_free()
	        spin_lock_irqsave()
	          rtlock_lock() --->  TRIGGERS
	                              lockdep_assert(!current->pi_blocked_on);

Adjust the check in put_task_struct() to also consider pi_blocked_on before
calling __put_task_struct(), resorting to the deferred call in case it is
set.

Suggested-by: Crystal Wood <crwood@redhat.com>
Signed-off-by: Luis Claudio R. Goncalves <lgoncalv@redhat.com>
---
 include/linux/sched/task.h |   16 +++++++++++++---
 1 file changed, 13 insertions(+), 3 deletions(-)

diff --git a/include/linux/sched/task.h b/include/linux/sched/task.h
index 0f2aeb37bbb04..638114f66a4d7 100644
--- a/include/linux/sched/task.h
+++ b/include/linux/sched/task.h
@@ -130,14 +130,22 @@ extern void __put_task_struct_rcu_cb(struct rcu_head *rhp);
 
 static inline void put_task_struct(struct task_struct *t)
 {
+	bool defer = false;
+
 	if (!refcount_dec_and_test(&t->usage))
 		return;
 
 	/*
 	 * In !RT, it is always safe to call __put_task_struct().
 	 * Under RT, we can only call it in preemptible context.
 	 */
-	if (!IS_ENABLED(CONFIG_PREEMPT_RT) || preemptible()) {
+#ifdef CONFIG_PREEMPT_RT
+	if (!preemptible() || current->pi_blocked_on)
+		defer = true;
+#endif
+
+	if (!defer) {
 		static DEFINE_WAIT_OVERRIDE_MAP(put_task_map, LD_WAIT_SLEEP);
 
 		lock_map_acquire_try(&put_task_map);
@@ -149,7 +156,9 @@ static inline void put_task_struct(struct task_struct *t)
 	/*
 	 * under PREEMPT_RT, we can't call put_task_struct
 	 * in atomic context because it will indirectly
-	 * acquire sleeping locks.
+	 * acquire sleeping locks. The same is true if the
+	 * current process has a mutex enqueued (blocked on
+	 * a PI chain).
 	 *
 	 * call_rcu() will schedule delayed_put_task_struct_rcu()
 	 * to be called in process context.
@@ -130,14 +130,22 @@ extern void __put_task_struct_rcu_cb(struct rcu_head *rhp);
 	/*
 	 * In !RT, it is always safe to call __put_task_struct().
-	 * Under RT, we can only call it in preemptible context.
+	 * Under RT, we can only call it in preemptible context,
+	 * when not blocked on a PI chain.
 	 */
-- 
2.49.0




^ permalink raw reply	[flat|nested] 5+ messages in thread

end of thread, other threads:[~2025-04-09 17:02 UTC | newest]

Thread overview: 5+ messages (download: mbox.gz / follow: Atom feed)
-- links below jump to the message on this page --
2025-04-09  0:27 [PATCH] sched: do not call __put_task_struct() on rt if pi_blocked_on is set Luis Claudio R. Goncalves
2025-04-09 15:43 ` Steven Rostedt
2025-04-09 16:27   ` Luis Claudio R. Goncalves
2025-04-09 16:39     ` Steven Rostedt
2025-04-09 17:02       ` Luis Claudio R. Goncalves

This is a public inbox, see mirroring instructions
for how to clone and mirror all data and code used for this inbox

all inboxes | Powered by JetHome®